/* 
 *	HT Editor
 *	htapp.cc
 *
 *	Copyright (C) 1999, 2000, 2001 Stefan Weyergraf (stefan@weyergraf.de)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License version 2 as
 *	published by the Free Software Foundation.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "cstream.h"
#include "htapp.h"
#include "htatom.h"
#include "htcfg.h"
#include "htclipboard.h"
#include "htdialog.h"
#include "hthelp.h"
#include "hthist.h"
#include "htidle.h"
#include "htinfo.h"
#include "htiobox.h"
#include "htkeyb.h"
#include "htmenu.h"
#include "htpal.h"
#include "htsearch.h"
#include "htstring.h"
#include "httree.h"
#include "htvfs.h"
#include "stream.h"
#include "tools.h"

#include "formats.h"

#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

extern "C" {
#include "regex.h"
}

#define ATOM_HT_APP		MAGICD("APP\0")

char _globalerror[1024];
char *globalerror=_globalerror;

/*
 *   app_stream_error_func()
 */

int app_stream_error_func(ht_stream *stream)
{
	int err=stream->get_error();
	char *name=stream->get_name();
	if (err & STERR_SYSTEM) {
		err=err&0xffff;
		switch (err) {
			case 4: {	/* EACCES*/
#ifdef DJGPP
				struct stat sbuf;

				stat(name, &sbuf);

				if (!(sbuf.st_mode & S_IWUSR)) {
					if (msgbox(msgbox_yes|msgbox_no, "title", 1, align_center, "%s: stream error (Permission denied), seems to be a (DOS) read-only file. Change attribute ?", name)==button_yes) {
						if (chmod(name, S_IRUSR | S_IWUSR)) {
							errorbox_modal("%s: error (%04x) changing attribute", name, errno & 0xffff);
						} else {
							stat(name, &sbuf);

							if (!(sbuf.st_mode & S_IWUSR)) {
								errorbox_modal("%s: error (%04x) changing attribute", name, errno & 0xffff);
							} else {
								return SERR_RETRY;
							}
						}
					}
				}

				break;
#endif
			}
			default:
				errorbox_modal("%s: stream error %04x: %s", name, err, strerror(err));
		}
	} else {
		err=err&0xffff;
		errorbox_modal("%s: internal stream error %04x", name, err);
	}
	return SERR_FAIL;
}

/*
 *   app_out_of_memory_proc()
 */
void *app_memory_reserve=0;

int app_out_of_memory_proc(int size)
{
	if (app_memory_reserve) {
		free(app_memory_reserve);
		app_memory_reserve = 0;
		warnbox_modal("the memory is getting low...");
		return OUT_OF_MEMORY_RETRY;
	} else {
		return OUT_OF_MEMORY_FAIL;
	}
}

/*
 *	CLASS ht_status
 */
void ht_status::init(bounds *b, char *Format)
{
	ht_view::init(b, VO_TRANSPARENT_CHARS, 0);
	VIEW_DEBUG_NAME("ht_status");
	idle_count = 0;
	analy_ani = 0;
	clear_len = 0;
	format = Format;
	render();
	register_idle_object(this);
}

void ht_status::done()
{
	unregister_idle_object(this);
	ht_view::done();
}

char *ht_status::defaultpalette()
{
	return palkey_menu_default;
}

char *ht_status::defaultpaletteclass()
{
	return palclasskey_menu;
}

void ht_status::draw()
{
	fill(size.w-clear_len, 0, clear_len, 1, getcolor(palidx_menu_normal), ' ');
	int len=strlen(workbuf);
	clear_len = len;
	buf_print(size.w-len, 0, getcolor(palidx_menu_normal), workbuf);
}

bool ht_status::idle()
{
	if (idle_count % 100 == 0) {
		char *oldstatus=ht_strdup(workbuf);
		render();
		if (strcmp(oldstatus, workbuf)) {
			dirtyview();
			redraw();
			screen->show();
		}
		
		analy_ani++;
		analy_ani &= 7;
	}
	idle_count++;
	return false;
}

void ht_status::render()
{
	strcpy(workbuf, format);
	char *buf = workbuf;
	char *f = format;
	if (f)
	while (*f) {
		if (*f==STATUS_ESCAPE) {
			switch (*(++f)) {
				case STATUS_ESCAPE:
					*(buf++) = STATUS_ESCAPE;
					break;
				case STATUS_ANALY_ACTIVE:
					if (some_analyser_active) {
						char *analysers[] = {"Analy", "aNaly", "anAly", "anaLy", "analY", "anaLy", "anAly", "aNaly"};
						strcpy(buf, analysers[analy_ani]);
						buf += 5;
					}
					break;
				case STATUS_ANALY_LINES:
					if (some_analyser_active) {
						buf += sprintf(buf, "(%d)", num_ops_parsed);
					}
					break;
				case STATUS_TIME: {
					time_t Time;
					time(&Time);
					tm *t=localtime(&Time);
					buf += sprintf(buf, "%02d:%02d", t->tm_hour, t->tm_min);
					break;
				}
				case STATUS_DATE: {
					time_t Time;
					time(&Time);
					tm *t=localtime(&Time);
					buf += sprintf(buf, "%02d.%02d.%04d", t->tm_mday, t->tm_mon+1, t->tm_year+1900);
					break;
				}
			}
		} else {
			*(buf++) = *f;
		}
		f++;
	}
	*buf = 0;
}


/*
 *	CLASS ht_keyline
 */

void ht_keyline::init(bounds *b)
{
	ht_view::init(b, 0, 0);
	VIEW_DEBUG_NAME("ht_keyline");
}

void ht_keyline::done()
{
	ht_view::done();
}

char *ht_keyline::defaultpalette()
{
	return palkey_keyline_default;
}

char *ht_keyline::defaultpaletteclass()
{
	return palclasskey_keyline;
}

void ht_keyline::draw()
{
	clear(getcolor(palidx_keyline_inactive));
	int x=0;
	for (int i=1; i<=10; i++) {
		htmsg msg;
		msg.type=mt_empty;
		msg.msg=msg_funcquery;
		msg.data1.integer=i;
		baseview->sendmsg(&msg);
		buf_printf(x, 0, getcolor(palidx_keyline_digit), "%d", i%10);
		if (msg.msg==msg_funcquery_ret) {
			char *s=msg.data1.str;
			if (s) {
				if (s[0]=='~') {
					buf_printf(x+1, 0, getcolor(palidx_keyline_inactive), s+1);
				} else {
					for (int j=0; j<size.w/10-1; j++) {
						buf_printf(x+j+1, 0, getcolor(palidx_keyline_active), " ");
					}
					buf_printf(x+1, 0, getcolor(palidx_keyline_active), s);
				}
			}
		}
		x+=size.w/10;
	}
}

void ht_keyline::handlemsg(htmsg *msg)
{
	ht_view::handlemsg(msg);
}

/*
 *	CLASS ht_desktop
 */

void ht_desktop::init(bounds *b)
{
	ht_view::init(b, VO_OWNBUFFER, 0);
	VIEW_DEBUG_NAME("ht_desktop");
}

void ht_desktop::done()
{
	ht_view::done();
}

char *ht_desktop::defaultpalette()
{
	return palkey_desktop_default;
}

char *ht_desktop::defaultpaletteclass()
{
	return palclasskey_desktop;
}

void ht_desktop::draw()
{
	fill(0, 0, size.w, size.h, getcolor(palidx_desktop_body), CHAR_FILLED_M);
}

/*
 *	CLASS ht_logviewer
 */

void ht_logviewer::init(bounds *b, int _maxlinecount)
{
	ht_viewer::init(b, "log", 0);
	VIEW_DEBUG_NAME("ht_logviewer");
	maxlinecount=_maxlinecount;
	lines=new ht_string_list();
	lines->init();
	ofs=0;
	xofs=0;
}

void ht_logviewer::done()
{
	lines->destroy();
	delete lines;
	ht_view::done();
}

int ht_logviewer::cursor_up(int n)
{
	ofs-=n;
	if (ofs<0) ofs=0;
	return n;
}

int ht_logviewer::cursor_down(int n)
{
	int c=lines->count();
	if (c>=size.h) {
		ofs+=n;
		if (ofs>c-size.h) ofs=lines->count()-size.h;
	}
	return n;
}

void ht_logviewer::deletefirstline()
{
	lines->del(0);
}

void ht_logviewer::draw()
{
	clear(getcolor(palidx_windialog_body));
	int c=lines->count();
	for (int i=0; i<size.h; i++) {
		if (i+ofs>=c) break;
		char *s=lines->get_string(i+ofs);
		int l=strlen(s);
		if (xofs<l) buf_print(0, i, getcolor(palidx_windialog_body), s+xofs);
	}
}

void ht_logviewer::insertline(char *line)
{
	int c=lines->count();
	if (c>=maxlinecount) deletefirstline();
	lines->insert_string(line);
	if (c>=size.h-1) {
		ofs=c-size.h+1;
	}
	dirtyview();
}

void ht_logviewer::handlemsg(htmsg *msg)
{
	if (msg->msg==msg_keypressed) {
		switch (msg->data1.integer) {
			case K_Up:
				cursor_up(1);
				dirtyview();
				clearmsg(msg);
				return;
			case K_Down:
				cursor_down(1);
				dirtyview();
				clearmsg(msg);
				return;
			case K_PageUp:
				cursor_up(size.h);
				dirtyview();
				clearmsg(msg);
				return;
			case K_PageDown:
				cursor_down(size.h);
				dirtyview();
				clearmsg(msg);
				return;
			case K_Right: case K_Control_Right:
				xofs+=2;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Left: case K_Control_Left:
				if (xofs-2>=0) xofs-=2;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Control_PageUp:
				ofs=0;
				dirtyview();
				clearmsg(msg);
				return;
			case K_Control_PageDown:
				ofs=lines->count()-size.h;
				if (ofs<0) ofs=0;
				dirtyview();
				clearmsg(msg);
				return;
		}
	}
	ht_viewer::handlemsg(msg);
}

/*
 *	CLASS ht_app_window_entry
 */

#define AWT_LOG		0
#define AWT_HELP		1
#define AWT_FILE		2
#define AWT_OFM		3

#define AWT_ISDESTROYABLE(type) (type!=AWT_LOG)

ht_app_window_entry::ht_app_window_entry(ht_window *w, UINT n, UINT t, bool m, void *d)
{
	window=w;
	number=n;
	type=t;
	minimized=m;
	data=d;
}

ht_app_window_entry::~ht_app_window_entry()
{
	if (AWT_ISDESTROYABLE(type)) {
		window->done();
		delete window;
	}
}

int compare_keys_app_window_entry(ht_data *key_a, ht_data *key_b)
{
	UINT a=((ht_app_window_entry*)key_a)->number;
	UINT b=((ht_app_window_entry*)key_b)->number;
	if (a>b) return 1; else if (a<b) return -1;
	return 0;
}

/*
 *	CLASS ht_app
 */

/*debug*/
//#define DRAW_TIMINGS
//#define NO_AVG
#define AVG_TIMINGS 10
int timings[AVG_TIMINGS];
int cur_timing=0, max_timing=0;
int h0;
/**/

void ht_app::init(bounds *_b)
{
	ht_dialog::init(_b, 0, 0);
	setframe(0);
	VIEW_DEBUG_NAME("ht_app");
	exit_program=0;
	focused=1;
	bounds b;

	windows=new ht_sorted_list();
	windows->init(compare_keys_app_window_entry);
	
/* init timer */
	h0=new_timer();
/* create menu */
	getbounds(&b);
	b.x=0;
	b.y=0;
	b.h=1;
	ht_menu *m=new ht_menu();
	m->init(&b);

	ht_static_context_menu *file=new ht_static_context_menu();
	file->init("~File");
	file->insert_entry("~Load", 0, cmd_file_load, 0, 1);
/*	file->insert_entry("~Save", 0, 0, 0, 0);
	file->insert_entry("~Save As...", 0, 0, 0, 0);*/
	file->insert_separator();
	file->insert_entry("~Quit", "Alt-X", cmd_quit, 0, 1);
	m->insert_menu(file);

	ht_static_context_menu *edit=new ht_static_context_menu();
	edit->init("~Edit");
	edit->insert_entry("~Copy", "Ctrl+Ins", cmd_clipboard_copy, 0, 1);
	edit->insert_entry("~Paste", "Shift+Ins", cmd_clipboard_paste, 0, 1);
	edit->insert_entry("S~how clipboard", 0, cmd_clipboard_show, 0, 1);
	edit->insert_entry("Cl~ear clipboard", 0, cmd_clipboard_clear, 0, 1);
	edit->insert_separator();
	edit->insert_entry("Copy ~from file...", 0, cmd_clipboard_copy_from_file, 0, 1);
	edit->insert_entry("Paste ~into file...", 0, cmd_clipboard_paste_into_file, 0, 1);
	m->insert_menu(edit);

	ht_static_context_menu *options=new ht_static_context_menu();
	options->init("~Options");
	options->insert_entry("~Palette", 0, cmd_options_palette, 0, 1);
	m->insert_menu(options);

	ht_static_context_menu *windows=new ht_static_context_menu();
	windows->init("~Windows");
	windows->insert_entry("~Size/Move", "Alt+F5", cmd_window_resizemove, K_Alt_F5, 1);
	windows->insert_entry("~Close", "Alt+F3", cmd_window_close, K_Alt_F3, 1);
	windows->insert_entry("~List", "Alt+0", cmd_window_list, K_Alt_0, 1);
	windows->insert_separator();
	windows->insert_entry("L~og window", 0, cmd_window_log, 0, 1);
	m->insert_menu(windows);

	ht_static_context_menu *help=new ht_static_context_menu();
	help->init("~Help");
	help->insert_entry("~Help", "F1", cmd_help_help, 0, 1);
	m->insert_menu(help);

	menu=m;
	insert(menu);
	
/* create status */
	/* the status should have the same bounds as the menu */
	ht_status *status = new ht_status();
	status->init(&b, STATUS_DEFAULT_FORMAT);
	insert(status);
	
/* create desktop */
	getbounds(&b);
	b.x=0;
	b.y=1;
	b.h-=2;
	desktop=new ht_desktop();
	desktop->init(&b);
	insert(desktop);
/* create keyline */
	getbounds(&b);
	b.x=0;
	b.y=b.h-1;
	b.h=1;
	keyline=new ht_keyline();
	keyline->init(&b);
	insert(keyline);
/* create battlefield */
	getbounds(&b);
	b.x=0;
	b.y=1;
	b.h-=2;

	battlefield=new ht_group();
	battlefield->init(&b, VO_TRANSPARENT_CHARS, "battlefield");
	insert(battlefield);

	logwindow=0;
	logviewer=0;
	helpwindow=0;
	
	create_log_window();
}

void ht_app::done()
{
	delete_timer(h0);

	if (windows) {
		windows->done();
		delete windows;
	}

	ht_dialog::done();
}

bool ht_app::create_log_window()
{
	if (!logwindow) {
		bounds b;
		battlefield->getbounds(&b);
		b.x=0;
		b.y=0;
		logwindow=new ht_window();
		logwindow->init(&b, "log window", FS_KILLER | FS_TITLE | FS_NUMBER | FS_RESIZE | FS_MOVE, 0);
		b.x=0;
		b.y=0;
		b.w-=2;
		b.h-=2;
		logviewer=new ht_logviewer();
		logviewer->init(&b, 128);
		logwindow->insert(logviewer);
	}
	
	ht_window *w=get_window_by_type(AWT_LOG);
	if (w) {
		focus(w);
	} else {		
		insert_window(logwindow, AWT_LOG, 0, 0);
	}
	return 1;
}

bool ht_app::create_help_window(char *file, char *node)
{
	ht_window *w=get_window_by_type(AWT_HELP);
	if (w) {
		focus(w);
		return 1;
	} else {
		bounds b, c;
		app->getbounds(&c);
		b.w=c.w*5/6;
		b.h=c.h*5/6;
		b.x=(c.w-b.w)/2;
		b.y=(c.h-b.h)/2;

		ht_window *window=new ht_window();
		window->init(&b, "help", FS_KILLER | FS_TITLE | FS_RESIZE | FS_MOVE | FS_NUMBER, 0);

		b.x=0;
		b.y=0;
		b.w-=2;
		b.h-=2;
		c=b;

		b=c;
		b.x+=b.w;
		b.w=1;
		ht_scrollbar *scrollbar=new ht_scrollbar();
		scrollbar->init(&b, &window->pal);
		scrollbar->enable();
		window->insert(scrollbar);

		b=c;
		ht_helpviewer *helpviewer=new ht_helpviewer();
		helpviewer->init(&b, scrollbar);
		window->insert(helpviewer);

		if (helpviewer->gotonode(file, node)) {
			helpwindow=window;

			insert_window(window, AWT_HELP, 0, 0);
			
			window->setpalette(palkey_windialog_cyan);

			return 1;
		}
		errorbox("help topic '(%s)%s' not found", file, node);
		window->done();
		delete window;
	}
	return 0;
}

bool ht_app::create_file_window(char *filename)
{
	bounds b;
	battlefield->getbounds(&b);
	b.x=0;
	b.y=0;
	char fullfilename[FILENAME_MAX];
	if (sys_canonicalize(filename, fullfilename)!=0) {
		LOG("invalid filename '%s'", filename);
		return 0;
	}
	ht_file *emfile=new ht_file();
	emfile->init(fullfilename, FAM_READ);
	ht_cached_file_pbd *file=new ht_cached_file_pbd();

	bool stream_cache_read=0;
	   
	file->init(emfile, 1, 256*1024,
		(stream_cache_read ? CF_CACHE_READS : 0) | CF_CACHE_WRITES |
			 CF_ONLY_EXPLICIT_FLUSHES, 8*1024);
	if (file->get_error()) {
		LOG("error loading file %s", fullfilename);
		return 0;
	}
	LOG("loading file %s...", fullfilename);
	file->set_error_func(app_stream_error_func);
	ht_window *window=new ht_window();
	window->init(&b, fullfilename, FS_KILLER | FS_TITLE | FS_NUMBER | FS_RESIZE | FS_MOVE, 0);
	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;
	ht_format_group *format_group=new ht_format_group();
	format_group->init(&b, VO_SELECTABLE | VO_RESIZE, VIEWERGROUP_NAME, file, true, true, format_viewer_ifs, 0);

	window->insert(format_group);
	LOG("%s: done.", fullfilename);
	
	insert_window(window, AWT_FILE, 0, file);
	return 1;
}

bool ht_app::create_ofm_window(char *url)
{
	bounds b;
	battlefield->getbounds(&b);
	b.x=0;
	b.y=0;
	ht_window *window=new ht_window();
	window->init(&b, "file manager", FS_KILLER | FS_TITLE | FS_NUMBER | FS_RESIZE | FS_MOVE, 0);

	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;
	ht_group *g=new ht_group();
	g->init(&b, 0, 0);

	bounds c=b;
	   
	b.x=0;
	b.y=0;
	b.h=1;
	ht_vfs_viewer_status *vst=new ht_vfs_viewer_status();
	vst->init(&b);
	   
	c.x=0;
	c.y=1;
	c.h--;
	ht_vfs_viewer *v=new ht_vfs_viewer();
	v->init(&c, "vfs viewer", 0, 0, 0, vst);
	   
	g->insert(v);
	   
	g->insert(vst);

/* build vfs list */
	ht_clist *vfslist=new ht_clist();
	vfslist->init();

/* file_vfs */
	ht_file_vfs *file_vfs=new ht_file_vfs();
	file_vfs->init();

/* reg_vfs */
	ht_reg_vfs *reg_vfs=new ht_reg_vfs();
	reg_vfs->init(registry);

	vfslist->insert(file_vfs);
	vfslist->insert(reg_vfs);
/**/

	ht_vfs_sub *vs=new ht_vfs_sub();
	vs->init(vfslist, url);

	v->insertsub(vs);

	v->sendmsg(msg_complete_init, 0);

	window->insert(g);
	
	insert_window(window, AWT_OFM, 0, 0);
	return 1;
}

char *ht_app::defaultpalette()
{
	return 0;
}

char *ht_app::defaultpaletteclass()
{
	return 0;
}

int analy_id = 0;
void ht_app::draw()
{
/* show draw timings */
#ifdef DRAW_TIMING
	int xyz=get_timer_1024tick(h0);
	buf->printf(17, 1, 7, "cur: %d (%d msec)", xyz*1024, get_timer_msec(h0));
#ifndef NO_AVG
	if (cur_timing>=AVG_TIMINGS-1) cur_timing=0;
	timings[cur_timing++]=xyz;
	if (cur_timing>max_timing) max_timing=cur_timing;
	int avg=0;
	for (int i=0; i<max_timing; i++) avg+=timings[i];
	avg=avg/max_timing;
	buf->printf(57, 1, 7, "avg%d: %d", max_timing+1, avg*1024);
#endif
#endif
/* display flag if analyser(s) active */
/*	char *analysers[] = {"Analy", "aNaly", "anAly", "anaLy", "analY", "anaLy", "anAly", "aNaly"};
	if (some_analyser_active) buf->printf(60, 0, VCP(VC_TRANSPARENT, VC_TRANSPARENT), "%s (%d)", analysers[analy_id++], num_ops_parsed);
	if (analy_id > 7) analy_id = 0;*/
}

void ht_app::delete_window(ht_window *window)
{
	UINT i=get_window_listindex(window);
	if (i!=LIST_UNDEFINED) {
		battlefield->remove(window);
//          windows->get(i);
		windows->del(i);
	}
}

UINT ht_app::find_free_window_number()
{
	UINT c=windows->count();
	UINT k=0;
repeat:
	k++;
	for (UINT i=0; i<c; i++) {
		ht_app_window_entry *e=(ht_app_window_entry*)windows->get(i);
		if (e->number==k) goto repeat;
	}
	return k;
}

void	ht_app::focus_window(ht_window *window)
{
	focus(window);
}

char *ht_app::func(UINT i, bool execute)
{
	switch (i) {
		case 1:
			if (execute) sendmsg(cmd_help_help);
			return "help";
		case 10:
			if (execute) sendmsg(cmd_quit);
			return "quit";
	}
	return 0;
}

ht_window *ht_app::get_window_by_number(UINT number)
{
	UINT c=windows->count();
	for (UINT i=0; i<c; i++) {
		ht_app_window_entry *e=(ht_app_window_entry*)windows->get(i);
		if (e->number==number) return e->window;
	}
	return 0;
}

ht_window *ht_app::get_window_by_type(UINT type)
{
	UINT c=windows->count();
	for (UINT i=0; i<c; i++) {
		ht_app_window_entry *e=(ht_app_window_entry*)windows->get(i);
		if (e->type==type) return e->window;
	}
	return 0;
}

UINT ht_app::get_window_number(ht_window *window)
{
	ht_app_window_entry *e=(ht_app_window_entry*)windows->get(get_window_listindex(window));
	if (e) return e->number; else return 0;
}

UINT ht_app::get_window_listindex(ht_window *window)
{
	UINT c=windows->count();
	for (UINT i=0; i<c; i++) {
		ht_app_window_entry *e=(ht_app_window_entry*)windows->get(i);
		if (e->window==window) return i;
	}
	return LIST_UNDEFINED;
}

void ht_app::handlemsg(htmsg *msg)
{
	if (msg->msg==msg_draw) {
		start_timer(h0);
		if (msg->type==mt_broadcast) {
			ht_view *v=first;
			while (v) {
				v->handlemsg(msg);
				v=v->next;
			}
		} else {
			current->handlemsg(msg);
		}
		stop_timer(h0);
		draw();
		screen->show();
	} else {
		ht_group::handlemsg(msg);
	}
	switch (msg->msg) {
		case msg_keypressed: {
			int i=0;
			switch (msg->data1.integer) {
				case K_Alt_9: i++;
				case K_Alt_8: i++;
				case K_Alt_7: i++;
				case K_Alt_6: i++;
				case K_Alt_5: i++;
				case K_Alt_4: i++;
				case K_Alt_3: i++;
				case K_Alt_2: i++;
				case K_Alt_1: i++;
					focus_window(get_window_by_number(i));
					clearmsg(msg);
					return;
				case K_F12: i++;
				case K_F11: i++;
				case K_F10: i++;
				case K_F9: i++;
				case K_F8: i++;
				case K_F7: i++;
				case K_F6: i++;
				case K_F5: i++;
				case K_F4: i++;
				case K_F3: i++;
				case K_F2: i++;
				case K_F1:
					i++;
					htmsg m;
					m.msg=msg_funcquery;
					m.type=mt_empty;
					m.data1.integer=i;
					sendmsg(&m);
					if (m.msg==msg_funcquery_ret) {
						sendmsg(msg_funcexec, i);
						clearmsg(msg);
						return;
					}
					break;
				case K_Alt_T:
					create_ofm_window("file:///");
					clearmsg(msg);
					return;
				case K_Alt_X:
					sendmsg(cmd_quit);
					clearmsg(msg);
					return;
				case K_Return:
					popup_view_list();
					clearmsg(msg);
					return;
				case K_Space: {
/*					msgbox(0, "title", 0, align_custom,
"ssssssssssssssssssssspppqqqrrrssstttuuu- xxx");*/
				msgbox(0, "title", 0, align_custom, "test, baby ! aaaaaaaaaaaargh,bf\n\n"
"Was willst DU Ich komm   nicht dahinter Ich zaehm dir das  Feuer      Du schenkst mir "
"nur Winter Wenn du nicht bald anfngst mir Antwort zu geben faengt es ohne "
"dich an das richtige Leben Ich hab lange gewartet bin nicht durchgeknallt "
"aber wenn du jetzt nicht \n\n\ecaufspringst wirst du ohne mich alt Erklaer mir ich "
"warte Wohin geht die Reise \n\n\erich kenn doch die Karte Nimm mich nicht als "
"Vorwand nimm mich nicht auf die Schippe sonst kuess ich dich luftleer und "
"zerbeiss dir die Lippe ich kann ohne dich leben auch wenn du nicht glaubst     \n"
"da ist mehr an mir dran als was du mir raubst ");
					break;
				}
			}
			break;
		}
		case msg_funcexec:
			if (func(msg->data1.integer, 1)) {
				clearmsg(msg);
				return;
			}
			break;
		case msg_funcquery: {
			char *s=func(msg->data1.integer, 0);
			if (s) {
				msg->msg=msg_funcquery_ret;
				msg->data1.str=s;
			} else clearmsg(msg);
			return;
		}
		case msg_truncate_file: {
			ht_streamfile *f=(ht_streamfile *)msg->data1.ptr;
			FILEOFS o=(FILEOFS)msg->data2.integer;
/*               ht_app_window_entry *e;
			if ((e=windows->enum_first())) {
				do {
					if ((e->type==AWT_FILE) && ((ht_streamfile*)e->data==f)) {
						check_collide();
					}
				} while ((e=windows->enum_first()));
			}*/
			if (confirmbox("really truncate %s at offset %08x ?", f->get_name(), o)==button_ok) {
				f->truncate(o);
			}
			clearmsg(msg);
			return;
		}
		case msg_kill:
			delete_window((ht_window*)msg->data1.ptr);
			clearmsg(msg);
			return;
		case cmd_quit:
			LOG("terminating...");
			exit_program=1;
			clearmsg(msg);
			return;
		case cmd_file_load: {
			char filename[129];
			filename[0]=0;
			if (inputbox("load file", "filename", filename, 128, HISTATOM_FILE)==button_ok) {
				if (filename[0]) create_file_window(filename);
			}
			clearmsg(msg);
			return;
		}
		case cmd_clipboard_show: {
			popup_clipboard_viewer();
			clearmsg(msg);
			return;
		}
		case cmd_clipboard_clear: {
			if (confirmbox("Do you really want to delete the clipboard ?")==button_ok) {
				clipboard_clear();
			}
			clearmsg(msg);
			return;
		}
		case cmd_clipboard_paste_into_file: {
			char filename[129];
			filename[0]=0;
			if (inputbox("clipboard - paste into file", "filename", filename, 128, HISTATOM_FILE)==button_ok) {
				ht_file *f=new ht_file();
				f->init(filename, FAM_WRITE | FAM_CREATE);
				if (!f->get_error()) {
					clipboard_paste(f, 0);
				} else errorbox("can't open file '%s'", filename);
				f->done();
				delete f;
			}
			clearmsg(msg);
			return;
		}
		case cmd_clipboard_copy_from_file: {
			char filename[129];
			filename[0]=0;
			if (inputbox("clipboard - copy from file", "filename", filename, 128, HISTATOM_FILE)==button_ok) {
				char desc[1024]; /* FIXME: possible buffer overflow */
				ht_file *f=new ht_file();
				f->init(filename, FAM_READ);
				if (!f->get_error()) {
					sprintf(desc, "file %s", f->get_name());
					clipboard_copy(desc, f, 0, f->get_size());
				} else errorbox("can't open file '%s'", filename);
				f->done();
				delete f;
			}
			clearmsg(msg);
			return;
		}
		case cmd_options_palette:
			create_ofm_window("reg:///palette");
			clearmsg(msg);
			return;
		case cmd_window_resizemove:
			sendmsg(msg_resize, 0);
			clearmsg(msg);
			return;
		case cmd_window_close:
			sendmsg(msg_kill, battlefield->current);
			clearmsg(msg);
			return;
		case cmd_window_list:
			popup_window_list();
			clearmsg(msg);
			return;
		case cmd_window_log:
			create_log_window();
			clearmsg(msg);
			return;
		case cmd_help_help:
			create_help_window("htinfo", "Top");
			clearmsg(msg);
			return;
	}
}

void	ht_app::insert_window(ht_window *window, UINT type, bool minimized, void *data)
{
	UINT n=find_free_window_number();
	ht_app_window_entry *e=new ht_app_window_entry(window, n, type, minimized, data);
	windows->insert(e);
	window->setnumber(n);
	battlefield->insert(window);
	focus(window);
}

int ht_app::load(ht_object_stream *f)
{
	ht_registry *temp;
	
	if (!(temp=(ht_registry*)f->get_object(NULL))) return 1;

	if (registry) {
		registry->done();
		delete registry;
	}
	registry=temp;
	
	load_history(f);
	
	htmsg m;
	m.msg=msg_config_changed;
	m.type=mt_broadcast;
	app->sendmsg(&m);
	return f->get_error();
}

void ht_app::log(char *line)
{
	logviewer->insertline(line);
	sendmsg(msg_draw);
}

void ht_app::logf(char *lineformat, ...)
{
	char buf[1024];	/* FIXME: possible buffer overflow ! */
	va_list arg;
	va_start(arg, lineformat);
	vsprintf(buf, lineformat, arg);
	va_end(arg);
	log(buf);
}

OBJECT_ID ht_app::object_id()
{
	return ATOM_HT_APP;
}

void	ht_app::popup_clipboard_viewer()
{
	bounds b;
	b.w=screen->size.w-4;
	b.h=screen->size.h*7/8;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;

	ht_dialog *d=new ht_dialog();
	d->init(&b, "clipboard", FS_TITLE | FS_KILLER | FS_MOVE | FS_RESIZE);

	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;
	ht_clipboard_viewer *v=new ht_clipboard_viewer();
	v->init(&b, "clipboard", VC_EDIT | VC_GOTO | VC_SEARCH, clipboard, 0);

	d->insert(v);
	d->setpalette(palkey_windialog_window_default);

	d->run(0);

	FILEOFS s, e;
	v->pselect_get(&s, &e);
	if (e>s) {
		clipboard->clipboard_start=s;
		clipboard->clipboard_size=e-s;
	} else {
		clipboard->clipboard_start=0;
		clipboard->clipboard_size=0;
	}

	d->done();
	delete d;
}

#if 1

int my_compare_func(char *a, char *b)
{
	return strcmp(a, b);
}

void ht_app::popup_view_list()
{
	if (!battlefield->current) return;
	bounds b, c;
	getbounds(&b);
	b.x=b.w/4;
	b.y=b.h/4;
	b.w/=2;
	b.h/=2;
	ht_dialog *dialog=new ht_dialog();
	dialog->init(&b, "Select mode", FS_KILLER | FS_TITLE | FS_MOVE);

/* create listbox */
	c=b;
	c.x=0;
	c.y=0;
	c.w-=2;
	c.h-=2;
	ht_text_listbox *listbox=new ht_text_listbox();
	listbox->init(&c);
/* insert all browsable views */
	ht_clist *structure=new ht_clist();
	((ht_clist*)structure)->init();
	int index=0;
	popup_view_list_dump(battlefield->current, listbox, structure, 0, &index, battlefield->getselected());
/* and seek to the currently selected one */
	listbox->update();
	listbox->seek(index);

	ht_text_listbox_sort_order so[1];
	so[0].col = 0;
	so[0].compare_func = my_compare_func;
/*     so[1].col = 1;
	so[1].compare_func = my_compare_func;*/
	
//     listbox->sort(1, so);

	dialog->insert(listbox);
	dialog->setpalette(palkey_windialog_special);

	if (dialog->run(0)) {
		ht_listbox_data data;
		listbox->databuf_get(&data);
		ht_data_ptr *p=(ht_data_ptr*)structure->get(data.cursor_id);
		focus((ht_view*)p->value);
	}

	structure->destroy();
	delete structure;

	dialog->done();
	delete dialog;
}

void ht_app::popup_view_list_dump(ht_view *view, ht_text_listbox *listbox, ht_list *structure, int depth, int *currenti, ht_view *currentv)
{
	if (!view) return;
	char str[256];
	char *s=str;
	if (!(view->options & VO_BROWSABLE)) {
		depth--;
	}

	for (int i=0; i<depth; i++) { *(s++)=' ';*(s++)=' '; }
	*s=0;

	if (!view) return;
	int c=view->childcount();
	for (int i=0; i<c; i++) {
		ht_view *v=view->getfirstchild();
		while (v) {
			if (v->browse_idx==i) break;
			v=v->next;
		}
		if (!v) return;
		
// FIXME: "viewergroup": dirty hack !!!
		if ((v->desc) && (strcmp(v->desc, "viewergroup")) && (v->options & VO_BROWSABLE)) {
			sprintf(s, "- %s", v->desc);
			structure->insert(new ht_data_ptr(v));
			if (v==currentv)
				*currenti=structure->count()-1;
			listbox->insert_str(structure->count()-1, str);
		}
		popup_view_list_dump(v, listbox, structure, depth+1, currenti, currentv);
	}
}

#else

void ht_app::popup_view_list()
{
	bounds b, c;
	getbounds(&b);
	b.x=b.w/4;
	b.y=b.h/4;
	b.w/=2;
	b.h/=2;
	ht_dialog *dialog=new ht_dialog(&b, "views");

/* create listbox */
	c=b;
	c.x=0;
	c.y=0;
	c.w-=2;
	c.h-=2;
	ht_static_treeview *treeview=new ht_static_treeview(&c, "jo");
/* insert all browsable views */
//	ht_collection *structure=new ht_collection();
//	int index=0;
	popup_view_list_dump(treeview, current, 0, 0);
/* and seek to the currently selected one */
//	listbox->seek(index);

//	treeview->add_child(0, "ertg");

	dialog->insert(treeview);
	dialog->setpalette(windialog_special_palette_str);

	if (execute_dialog(dialog)) {
/*		ht_listbox_data data;
		listbox->getdata(&data);
		ht_data_ptr *p=(ht_data_ptr*)structure->get(data.index);
		focus((ht_view*)p->data);*/
	}
	dialog->done();
	delete dialog;
}

void ht_app::popup_view_list_dump(ht_static_treeview *treeview, ht_view *view, void *node, int depth)
{
	void *r;
	if ((view->options & VO_BROWSABLE) && (depth)) {
		char str[128];
		strcpy(str, view->desc);
		r=treeview->add_child(node, str);
	} else r=node;

	int c=view->childcount();
	for (int i=0; i<c; i++) {
		ht_view *v=view->getfirstchild();
		while (v) {
			if (v->browse_idx==i) break;
			v=v->next;
		}
		if (!v) return;
//		if (v->options & VO_BROWSABLE) {
			popup_view_list_dump(treeview, v, r, depth++);
//		}
	}
}
#endif


void ht_app::popup_window_list()
{
	bounds b, c;
	getbounds(&b);
	b.x=b.w/4;
	b.y=b.h/4;
	b.w/=2;
	b.h/=2;
	ht_dialog *dialog=new ht_dialog();
	dialog->init(&b, "windows", FS_KILLER | FS_TITLE | FS_MOVE);

/* create listbox */
	c=b;
	c.x=0;
	c.y=0;
	c.w-=2;
	c.h-=2;
	ht_text_listbox *listbox=new ht_itext_listbox();
	listbox->init(&c, 2, 1);

	UINT vc=windows->count();
	for (UINT i=0; i<vc; i++) {
		ht_app_window_entry *e=(ht_app_window_entry*)windows->get(i);
		char l[16];
		sprintf(l, " %2d", e->number);
		listbox->insert_str(e->number, l, e->window->desc);
	}
	listbox->update();
	if (battlefield->current) listbox->seek(battlefield->current->getnumber()-1);

	dialog->insert(listbox);
	dialog->setpalette(palkey_windialog_special);

	if (dialog->run(0)) {
		ht_listbox_data data;
		listbox->databuf_get(&data);
		focus_window(get_window_by_number(data.cursor_id));
	}
	dialog->done();
	delete dialog;
}

int ht_app::run(bool modal)
{
	sendmsg(msg_draw, 0);
	while (!exit_program) {
		if (ht_keypressed()) {
			sendmsg(msg_keypressed, ht_getkey());
			sendmsg(msg_draw);
		}
		ht_queued_msg *q;
		while ((q=dequeuemsg())) {
			htmsg m=q->msg;
			q->target->sendmsg(&m);
			sendmsg(msg_draw);
		}
		do_idle();
	}
	return 0;
}

void ht_app::store(ht_object_stream *f)
{
	f->put_object(registry, 0);

	store_history(f);

	return;
	
	f->put_int_dec(windows->count(), 4, "windows");
	ht_app_window_entry *e;
	if ((e=(ht_app_window_entry*)windows->enum_first())) {
		char prjcfg_path[1024];	/* FIXME: possible buffer overflow */
		sys_dirname(f->get_name(), prjcfg_path);
		do {
			f->put_int_dec(e->type, 4, "type");
			switch (e->type) {
				case AWT_LOG:
					break;
				case AWT_HELP:
					break;
				case AWT_FILE: {
					char relfilename[1024];	/* FIXME: possible buffer overflow */
					sys_relname(e->window->desc, prjcfg_path, relfilename);
					strcat(relfilename, ".htcfg");
					f->put_string(relfilename, "file");
					save_fileconfig(relfilename, e->window);
					break;
				}
				case AWT_OFM:
					break;
			}
		} while ((e=(ht_app_window_entry*)windows->enum_next()));
	}
}

/*
 *
 */


void insert_search_history_entry(ht_clist *hist, ht_view *form, byte *str, UINT slen, byte *start, UINT startlen, byte *end, UINT endlen, UINT state)
{
	char s[1024], *tt=s;	/* FIXME: possible buffer overflow */
	tt+=escape_special(tt, str, slen, NULL, false);
	strcat(tt, " (");
	if (state & 1) {
		strcat(tt, "case-insensitive");
	} else {
		strcat(tt, "case-sensitive");
	}
	if (startlen) {
		strcat(tt, ", start=");
		tt+=strlen(tt);
		tt+=escape_special(tt, start, startlen);
	}
	if (endlen) {
		strcat(tt, ", end=");
		tt+=strlen(tt);
		tt+=escape_special(tt, end, endlen);
	}
	strcat(tt, ")");
	insert_history_entry(hist, s, form);
}

bool test_str_to_ofs(FILEOFS *ofs, byte *str, UINT strlen, ht_format_viewer *format, char *desc)
{
#define TEST_STR_TO_OFS_MAXSTRLEN		128
	if (strlen>TEST_STR_TO_OFS_MAXSTRLEN) {
		errorbox("%s: expression too long (len %d, max %d)", desc, strlen, TEST_STR_TO_OFS_MAXSTRLEN);
		return false;
	}
	if (strlen>0) {
		char s[TEST_STR_TO_OFS_MAXSTRLEN+1];
		bin2str(s, str, strlen);
		if (!format->string_to_offset(s, ofs)) {
			errorbox("%s: invalid expression: '%s'", desc, s);
			return false;
		}
	}
	return true;
}

bool test_str_to_addr(fmt_vaddress *addr, byte *str, UINT strlen, ht_format_viewer *format, char *desc)
{
#define TEST_STR_TO_ADDR_MAXSTRLEN		128
	if (strlen>TEST_STR_TO_ADDR_MAXSTRLEN) {
		errorbox("%s: expression too long (len %d, max %d)", desc, strlen, TEST_STR_TO_ADDR_MAXSTRLEN);
		return false;
	}
	if (strlen>0) {
		char s[TEST_STR_TO_ADDR_MAXSTRLEN+1];
		bin2str(s, str, strlen);
		if (!format->string_to_address(s, addr)) {
			errorbox("%s: invalid expression: '%s'", desc, s);
			return false;
		}
	}
	return true;
}

/* FIXME: global var */
UINT lastsearchmodeid=0;

void search_dialog(ht_format_viewer *format, UINT searchmodes)
{
	if (!searchmodes) return;
	
	bounds b;
	b.w=50;
	b.h=15;
	b.x=(screen->size.w-b.w)/2;
	b.y=(screen->size.h-b.h)/2;
	ht_search_dialog *dialog=new ht_search_dialog();
	dialog->init(&b);
	b.x=0;
	b.y=0;
	b.w-=2;
	b.h-=2;

	int dofs=0;

	ht_clist *hist_hexascii=0;
	if (searchmodes & SEARCHMODE_BIN) {
		ht_hexascii_search_mask *s0=new ht_hexascii_search_mask();
		hist_hexascii=(ht_clist*)find_atom(HISTATOM_SEARCH_BIN);
		s0->init(&b, 0, hist_hexascii);
		dialog->insertmode(0, "bin: ascii/hex", s0);
		dofs+=dialog->datasize();
	}

	ht_clist *hist_casm=0;
	if (searchmodes & SEARCHMODE_CASM) {
		ht_casm_search_mask *s1=new ht_casm_search_mask();
		hist_casm=(ht_clist*)find_atom(HISTATOM_SEARCH_CASM);
		s1->init(&b, 0, hist_casm);
		dialog->insertmode(1, "bin: c/asm", s1);
		dofs+=dialog->datasize();
	}

	ht_clist *hist_text=0;
	if (searchmodes & SEARCHMODE_TEXT) {
		ht_text_search_mask *s1=new ht_text_search_mask();
		hist_text=(ht_clist*)find_atom(HISTATOM_SEARCH_TEXT);
		s1->init(&b, 0, hist_text);
		dialog->insertmode(2, "display: regex", s1);
		dofs+=dialog->datasize();
	}

	dialog->selectmode(lastsearchmodeid);

	if (dialog->run(0)) {
		ht_search_dialog_data *t=(ht_search_dialog_data *)malloc(dialog->datasize());
		dialog->databuf_get(t);

		int modeid=t->modes_id;

		free(t);

		ht_view *form=dialog->getmodeform();
		UINT ds=form->datasize();
		void *formdata=malloc(ds);
		form->databuf_get(formdata);
		
		lastsearchmodeid=modeid;
		
		switch (modeid) {
			case 0: {
/* ascii/hex */
				ht_hexascii_search_mask_data *d=(ht_hexascii_search_mask_data*)formdata;
				if (d->str.pure.textlen) {
					insert_search_history_entry(hist_hexascii, form, d->str.pure.text, d->str.pure.textlen, d->start.pure.text, d->start.pure.textlen, d->end.pure.text, d->end.pure.textlen, d->options.state);

					FILEOFS start=0, end=0xffffffff;
					format->get_current_offset(&start);

					if (test_str_to_ofs(&start, d->start.pure.text, d->start.pure.textlen, format, "start-offset")
					&& test_str_to_ofs(&end, d->end.pure.text, d->end.pure.textlen, format, "end-offset")) {
//						infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
						ht_fxbin_search_request *q=new ht_fxbin_search_request(SC_PHYSICAL, d->options.state & 1 ? SF_FXBIN_CASEINSENSITIVE: 0, d->str.pure.textlen, d->str.pure.text);
						ht_physical_search_result *r=(ht_physical_search_result*)format->psearch(q, start, end);
						if (r) {
							format->goto_offset(r->offset, format);
							format->pselect_set(r->offset, r->offset+r->size);
							delete r;
						} else infobox("string not found");
						delete q;
					}
				}
				break;
			}
			case 1: {
/* C/ASM */
#define CASM_MAXSTRLEN		128
				ht_casm_search_mask_data *d=(ht_casm_search_mask_data*)formdata;
				if (d->str.pure.textlen) {
					insert_search_history_entry(hist_casm, form, d->str.pure.text, d->str.pure.textlen, d->start.pure.text, d->start.pure.textlen, d->end.pure.text, d->end.pure.textlen, d->options.state);

					char strbuf[CASM_MAXSTRLEN+1];
					if (d->str.pure.textlen<=CASM_MAXSTRLEN) {
						bin2str(strbuf, d->str.pure.text, d->str.pure.textlen);

						int slen;
						byte *s=eval_casmstr(strbuf, &slen);

						FILEOFS start=0, end=0xffffffff;
						format->get_current_offset(&start);

						if (s && slen) {
							if (test_str_to_ofs(&start, d->start.pure.text, d->start.pure.textlen, format, "start-offset")
							&& test_str_to_ofs(&end, d->end.pure.text, d->end.pure.textlen, format, "end-offset")) {
//								infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
								ht_fxbin_search_request *q=new ht_fxbin_search_request(SC_PHYSICAL, d->options.state & 1 ? SF_FXBIN_CASEINSENSITIVE: 0, slen, s);
								ht_physical_search_result *r=(ht_physical_search_result*)format->psearch(q, start, end);
								if (r) {
									format->goto_offset(r->offset, format);
									format->pselect_set(r->offset, r->offset+r->size);
									delete r;
								} else infobox("string not found");
								delete q;
							}
						} else errorbox("C/asm string: %s in '%s'", casmerrstr, s);
					} else errorbox("expression too long (size %d, max %d)", strlen, CASM_MAXSTRLEN);
				}
				break;
			}
			case 2: {
/* text */
				ht_text_search_mask_data *d=(ht_text_search_mask_data*)formdata;
				if (d->str.pure.textlen) {
					insert_search_history_entry(hist_text, form, d->str.pure.text, d->str.pure.textlen, d->start.pure.text, d->start.pure.textlen, d->end.pure.text, d->end.pure.textlen, d->options.state);

					char strbuf[CASM_MAXSTRLEN+1];
					if (d->str.pure.textlen<=CASM_MAXSTRLEN) {
						bin2str(strbuf, d->str.pure.text, d->str.pure.textlen);

						fmt_vaddress start=0, end=0xffffffff;
						format->get_current_address(&start);

						if (test_str_to_addr(&start, d->start.pure.text, d->start.pure.textlen, format, "start-address")
						&& test_str_to_addr(&end, d->end.pure.text, d->end.pure.textlen, format, "end-address")) {
//							infobox("debug: searching for '%s' from %08x to %08x", d->str, start, end);
							try {
								ht_regex_search_request *q=new ht_regex_search_request(SC_VISUAL, d->options.state & 1 ? SF_REGEX_CASEINSENSITIVE : 0, strbuf);
								ht_visual_search_result *r=(ht_visual_search_result *)format->vsearch(q, start, end);
								if (r) {
									format->goto_address(r->address, format);
									delete r;
								} else infobox("string not found");
								delete q;
							} catch (ht_regex_search_exception *r) {
								char rxerr[128];
								regerror(r->errorcode, r->regex, rxerr, 128);
								errorbox("error compiling regex: %s", rxerr);
							}
						}
					} else errorbox("expression too long (size %d, max %d)", strlen, CASM_MAXSTRLEN);
				}
				break;
			}
			default:
				warnbox("search mode (modeid=%d) not supported !", modeid);
		}
		free(formdata);
	}
	dialog->done();
	delete dialog;
}

BUILDER(ATOM_HT_APP, ht_app);

/*
 *	INIT
 */

bool init_app()
{
	bounds b;
	screen=new screendrawbuf(ht_name" "ht_version);

	b.x=0;
	b.y=0;
	b.w=screen->size.w;
	b.h=screen->size.h;
	app=new ht_app();
	((ht_app*)app)->init(&b);
	baseview=app;

	app_memory_reserve = malloc(16384); // malloc, not smalloc
	out_of_memory_func = &app_out_of_memory_proc;

	REGISTER(ATOM_HT_APP, ht_app);

	return true;
}

/*
 *	DONE
 */

void done_app()
{
	UNREGISTER(ATOM_HT_APP, ht_app);
	
	out_of_memory_func = &out_of_memory;
	if (app_memory_reserve) free(app_memory_reserve);

	if (app) {
		app->done();
		delete app;
	}
	
	if (screen) delete screen;
}

